#pragma once

#include "Utility/ProjectUtils.hpp"

#if PLATFORM_WINDOWS
#include "Windows/AllowWindowsPlatformTypes.h"
#endif
#include <Windows.h>
#include <stdio.h>
#include <conio.h>
#if PLATFORM_WINDOWS
#include "Windows/HideWindowsPlatformTypes.h"
#endif

#include "RHICommandList.h"

template <class T>
class SharedQueueWriter
{
private:
	const int KEY_SIZE = 256;

	int dataSize;

	HANDLE filledEvent = NULL;
	HANDLE emptiedEvent = NULL;

	HANDLE dataFile = NULL;
	void* dataPtr = NULL;

#if MIXCAST_D3D12
	bool useEngineFence = false;
	FGPUFenceRHIRef fence;
#endif

	ID3D11Device* fenceDev;
	ID3D11DeviceContext* fenceContext = NULL;
	ID3D11Query* fenceQuery = NULL;

	//render thread functions
	void InsertFilledFence(FRHICommandListImmediate& RHICmdList)
	{
#if MIXCAST_D3D12
		if (useEngineFence)
		{
			fence = RHICmdList.CreateGPUFence(FName("Queue Write"));
			RHICmdList.WriteGPUFence(fence);
		}
		else
		{
#endif
			D3D11_QUERY_DESC queryDesc;
			queryDesc.Query = D3D11_QUERY_EVENT;
			queryDesc.MiscFlags = 0;
			fenceDev->CreateQuery(&queryDesc, &fenceQuery);

			fenceDev->GetImmediateContext(&fenceContext);
			fenceContext->End(fenceQuery);
#if MIXCAST_D3D12
		}
#endif
	}
	void UpdateFence()
	{
#if MIXCAST_D3D12
		if (useEngineFence)
		{
			if (!fence.IsValid())
				return;

			if (fence->Poll())
				MarkFilled();
		}
		else
		{
#endif
			if (fenceQuery == NULL)
				return;

			if (fenceContext->GetData(fenceQuery, NULL, 0, 0) == S_OK)
				MarkFilled();
#if MIXCAST_D3D12
		}
#endif
	}

public:
	SharedQueueWriter(LPCWSTR frameId) 
	{
		this->dataSize = sizeof(T);

#if MIXCAST_D3D12
		useEngineFence = ProjectUtils::GetGraphicsType() != ExpGraphicsType::DX11;
		if (!useEngineFence)
#endif
			fenceDev = (ID3D11Device*)GDynamicRHI->RHIGetNativeDevice();

		LPWSTR key = new TCHAR[KEY_SIZE];
		//Events
		wsprintf(key, TEXT("Global\\MixCast.SharedQ(%s).Filled"), frameId);
		filledEvent = CreateEvent(NULL, true, false, key);
		ResetEvent(filledEvent);
		wsprintf(key, TEXT("Global\\MixCast.SharedQ(%s).Emptied"), frameId);
		emptiedEvent = CreateEvent(NULL, true, false, key);
		ResetEvent(emptiedEvent);

		//Metadata
		wsprintf(key, TEXT("MixCast-SharedQ(%s)-Data"), frameId);
		dataFile = CreateFileMapping(
			INVALID_HANDLE_VALUE,		// use paging file
			NULL,						// default security
			PAGE_READWRITE,				// read/write access
			0,							// maximum object size (high-order DWORD)
			dataSize,				// maximum object size (low-order DWORD)
			key);						// name of mapping object
		dataPtr = (void*)MapViewOfFile(dataFile,   // handle to map object
			FILE_MAP_ALL_ACCESS, // read/write permission
			0,
			0,
			dataSize);
		ZeroMemory(dataPtr, dataSize);

		delete[] key;
	}
	~SharedQueueWriter()
	{
#if MIXCAST_D3D12
		fence.SafeRelease();
#endif
		SAFE_RELEASE(fenceContext);
		SAFE_RELEASE(fenceQuery);

		UnmapViewOfFile(dataPtr);
		CloseHandle(dataFile);

		CloseHandle(filledEvent);
		CloseHandle(emptiedEvent);
	}

	bool WaitUntilEmptied(int timeout)
	{
		if (timeout == -1) {
			WaitForSingleObject(emptiedEvent, INFINITE);
			ResetEvent(emptiedEvent);
			return true;
		}
		else {
			DWORD res = WaitForSingleObject(emptiedEvent, timeout);
			if (res == WAIT_OBJECT_0)
			{
				ResetEvent(emptiedEvent);
				return true;
			}
			return false;
		}
	}

	void Write(T src, bool requiresGpuOp)
	{
		CopyMemory(dataPtr, &src, dataSize);
		if (!requiresGpuOp)
			MarkFilled();
		else
			EnqueueFenceInsert();
	}
	void MarkFilled()
	{
#if MIXCAST_D3D12
		fence.SafeRelease();
#endif
		SAFE_RELEASE(fenceContext);
		SAFE_RELEASE(fenceQuery);

		SetEvent(filledEvent);
	}
	void EnqueueFenceInsert()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
		{
			InsertFilledFence(RHICmdList);
		});
	}
	void EnqueueFenceUpdate()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
		{
			UpdateFence();
		});
	}
};