#pragma once

#include "OwnedUnrealSharedTexture_D3D11.h"
#include "WatchedUnrealSharedTexture.h"

#if PLATFORM_WINDOWS
#include "Windows/AllowWindowsPlatformTypes.h"
#endif
#include <d3d11.h>
#include <dxgi1_2.h>
#if PLATFORM_WINDOWS
#include "Windows/HideWindowsPlatformTypes.h"
#endif

#if MIXCAST_D3D12
#include "ITextureShareD3D11.h"
#endif

#include "../../../MixCastLog.h"

void OwnedUnrealSharedTexture_D3D11::Write_RenderThread(FRHICommandListImmediate& RHICmdList)
{
	if (sharedTex == NULL)
		return;

	ID3D11Texture2D* src = (ID3D11Texture2D*)GetNativeTexture(*srcTex);
	if (src == NULL)
		return;

	ID3D11DeviceContext* ctx = NULL;
	device->GetImmediateContext(&ctx);

	bool proceed = true;

	IDXGIKeyedMutex* mutex;
	HRESULT hr = sharedTex->QueryInterface<IDXGIKeyedMutex>(&mutex);
	if (SUCCEEDED(hr) && mutex)
		proceed = SUCCEEDED(mutex->AcquireSync(0, INFINITE));

	if (proceed) {
#if MIXCAST_D3D12
		FBox2D CopyBox(FVector2D(0.0f, 0.0f), FVector2D(srcTex->GetSurfaceWidth(), srcTex->GetSurfaceHeight()));
		GDynamicRHI->RHICopySubTextureRegion(srcTex->Resource->GetTexture2DRHI(), sharedTexRhi, CopyBox, CopyBox);
#else
		ctx->CopyResource(sharedTex, src);
#endif
		if (SUCCEEDED(hr) && mutex)
			mutex->ReleaseSync(0);
	}
	if (SUCCEEDED(hr) && mutex)
		mutex->Release();

	ctx->Release();
}

OwnedUnrealSharedTexture_D3D11::OwnedUnrealSharedTexture_D3D11(UTexture* srcTex)
{
	this->srcTex = srcTex;
	ID3D11Texture2D* src = (ID3D11Texture2D*)GetNativeTexture(*srcTex);
	src->GetDevice(&device);

	D3D11_TEXTURE2D_DESC srcDesc;
	src->GetDesc(&srcDesc);

	D3D11_TEXTURE2D_DESC newDesc;
	newDesc.Width = srcDesc.Width;
	newDesc.Height = srcDesc.Height;
	newDesc.MipLevels = 1;
	newDesc.ArraySize = 1;
	newDesc.Format = EnsureShareableFormat(srcDesc.Format);
	newDesc.Usage = D3D11_USAGE_DEFAULT;
	newDesc.SampleDesc.Count = 1;
	newDesc.SampleDesc.Quality = 0;
	newDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
	newDesc.CPUAccessFlags = 0;
	newDesc.MiscFlags = (D3D11_RESOURCE_MISC_SHARED_NTHANDLE | D3D11_RESOURCE_MISC_SHARED_KEYEDMUTEX);

	try
	{
		HRESULT hr = device->CreateTexture2D(&newDesc, NULL, &sharedTex);

		if (FAILED(hr)) {
			return;
		}
	}
	catch (...)
	{
		MIXCAST_ERROR("Failed to create shared texture");
		return;
	}

	try
	{
		HRESULT hr = sharedTex->QueryInterface(__uuidof(IDXGIResource), (void**)&dxgi_res);

		if (FAILED(hr)) {
			MIXCAST_ERROR("Failed to get dxgi resource");
			SAFE_RELEASE(sharedTex);
			return;
		}

		hr = ((IDXGIResource1*)dxgi_res)->CreateSharedHandle(NULL, DXGI_SHARED_RESOURCE_READ | DXGI_SHARED_RESOURCE_WRITE, NULL, &sharedHandle);

		if (FAILED(hr)) {
			MIXCAST_ERROR("Failed to create shared handle");
			SAFE_RELEASE(sharedTex);
			return;
		}
	}
	catch (...)
	{
		SAFE_RELEASE(dxgi_res);
		SAFE_RELEASE(sharedTex);
		sharedHandle = NULL;

		MIXCAST_ERROR("Failed to get shared handle");
		return;
	}

#if MIXCAST_D3D12
	if(!ITextureShareD3D11::Get().CreateRHITexture(sharedTex, WatchedUnrealSharedTexture::GetUnrealFormatFromDirectXFormat(newDesc.Format), sharedTexRhi))
		MIXCAST_ERROR("Couldn't create RHI for shared owned texture");
	sharedTexRhi->SetName(FName("Shared Tex"));
#endif
}
OwnedUnrealSharedTexture_D3D11::~OwnedUnrealSharedTexture_D3D11()
{
	if (sharedHandle == NULL)
		return;

#if MIXCAST_D3D12
	SAFE_RELEASE(sharedTexRhi);
#endif
	SAFE_RELEASE(dxgi_res);
	SAFE_RELEASE(sharedTex);
	SAFE_RELEASE(device);
}